use core::slice::{Iter, IterMut};
use std::ops::{Index, IndexMut};
use std::{fmt, fmt::Debug};

pub struct VkMappedMemory<'a, T>
where
    T: Clone,
{
    data: &'a mut [T],

    unmap: Option<Box<dyn Fn()>>,
}

impl<'a, T: Clone> VkMappedMemory<'a, T> {
    pub fn new(data: &'a mut [T]) -> VkMappedMemory<'a, T> {
        VkMappedMemory { data, unmap: None }
    }

    pub fn set_unmap<F>(&mut self, f: F)
    where
        F: Fn() + 'static,
    {
        self.unmap = Some(Box::new(f));
    }

    pub fn copy(&mut self, data: &[T]) {
        self.data.clone_from_slice(data);
    }

    pub fn iter(&self) -> Iter<'_, T> {
        self.data.iter()
    }

    pub fn iter_mut(&mut self) -> IterMut<'_, T> {
        self.data.iter_mut()
    }
}

impl<'a, T: Clone> Index<usize> for VkMappedMemory<'a, T> {
    type Output = T;

    fn index(&self, index: usize) -> &T {
        &self.data[index]
    }
}

impl<'a, T: Clone> IndexMut<usize> for VkMappedMemory<'a, T> {
    fn index_mut(&mut self, index: usize) -> &mut T {
        &mut self.data[index]
    }
}

impl<'a, T: Clone> Drop for VkMappedMemory<'a, T> {
    fn drop(&mut self) {
        if let Some(unmap) = &self.unmap {
            unmap();
        }
    }
}

impl<'a, T: Clone + Debug> fmt::Debug for VkMappedMemory<'a, T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("VkMappedMemory")
            .field("data", &self.data)
            .finish()
    }
}
